from abc import ABC, abstractmethod

from opoly.statements import Statement, StatementType, ForLoopStatement, DeclarationStatement, AssignmentStatement


class CodeGenerator(ABC):

    @abstractmethod
    def generate(self, stmt: Statement) -> str:
        pass


class PseudoCodeGenerator(CodeGenerator):

    INDENTATION_SPACES = " " * 4

    def generate_for_loop(self, stmt: ForLoopStatement, level=0) -> str:
        conc = f" CONC" if stmt._is_parallel else ""
        head = f"FOR{conc} {stmt.index} FROM {stmt.lowerbound} TO {stmt.upperbound} STEP {stmt.step} " + "{\n"
        body = f"\n".join([self.generate(t, level=level+1) for t in stmt.body])
        return (self.INDENTATION_SPACES * level) + head + body + "\n" + (self.INDENTATION_SPACES * level) + "}"

    def generate_declaration(self, stmt: DeclarationStatement, level=0) -> str:
        head = f"VAR {stmt.variable}"
        init_str = f" = {stmt.initialization}" if stmt.initialization is not None else ""
        return (self.INDENTATION_SPACES * level) + head + init_str + ";"

    def generate_assignment(self, stmt: AssignmentStatement, level=0) -> str:
        return (self.INDENTATION_SPACES * level) + f"STM {stmt.left_term} = {stmt.right_term};"

    def generate(self, stmt: Statement, level=0) -> str:
        if stmt.stype == StatementType.FOR_LOOP:
            return self.generate_for_loop(stmt, level=level)
        if stmt.stype == StatementType.DECLARATION:
            return self.generate_declaration(stmt, level=level)
        if stmt.stype == StatementType.ASSIGNMENT:
            return self.generate_assignment(stmt, level=level)
        return None

class CCodeGenerator(CodeGenerator):

    INDENTATION_SPACES = " " * 4

    def generate_for_loop(self, stmt: ForLoopStatement, level=0) -> str:
        omp_directive = (self.INDENTATION_SPACES * level) + "#pragma omp parallel for\n" if stmt.is_parallel else ""
        step = f"{stmt.index}++" if stmt.step.is_constant() and stmt.step.value == 1 else f"{stmt.index} += {stmt.step}"
        head = (self.INDENTATION_SPACES * level) + f"for(int {stmt.index} = {stmt.lowerbound}; {stmt.index} <= {stmt.upperbound}; {step}) " + "{\n"
        body = f"\n".join([self.generate(t, level=level+1) for t in stmt.body])
        return omp_directive + head + body + "\n" + (self.INDENTATION_SPACES * level) + "}"

    def generate_declaration(self, stmt: DeclarationStatement, level=0) -> str:
        head = f"{stmt.var_type} {stmt.variable}"
        init_str = f" = {stmt.initialization}" if stmt.initialization is not None else ""
        return (self.INDENTATION_SPACES * level) + head + init_str + ";"

    def generate_assignment(self, stmt: AssignmentStatement, level=0) -> str:
        return (self.INDENTATION_SPACES * level) + f"{stmt.left_term} = {stmt.right_term};"

    def generate(self, stmt: Statement, level=0) -> str:
        if stmt.stype == StatementType.FOR_LOOP:
            return self.generate_for_loop(stmt, level=level)
        if stmt.stype == StatementType.DECLARATION:
            return self.generate_declaration(stmt, level=level)
        if stmt.stype == StatementType.ASSIGNMENT:
            return self.generate_assignment(stmt, level=level)
        return None



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="opoly", # Replace with your own username
    version="0.1.3",
    author="Giacomo Aloisi (GiackAloZ)",
    author_email="giacomo.aloisi1998@gmail.com",
    description="OPoly: a simple OpenMP polyhedral compilator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GiackAloZ/OPoly",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Compilers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    keywords="compiler polyhedral optimization parallel omp c",
    packages=setuptools.find_packages(),
    package_data={
        "opoly.modules.minizinc": ["models/*.mzn", "libraries/*.mzn"],
    },
    install_requires=[
        "numpy>=1.19",
        "sympy>=1.7",
        "pymzn>=0.18.3"
    ],
    python_requires='>=3.9',
    entry_points={
        "console_scripts": [
            "opoly=opoly.scripts.opoly:main",
        ],
    }
)


import pytest

from opoly.expressions import (
    Expression,
    GroupingExpression,
    ConstantExpression,
    VariableExpression,
    FunctionExpression,
    UnaryExpression,
    extract_variable_expressions,
    divide_variable_expressions_by_name
)


class TestExpressions():

    def test_constant_expression(self):
        expr = ConstantExpression(1)
        assert str(expr) == "1"
        assert expr.is_constant()

    def test_simple_variable_expression(self):
        expr = VariableExpression("a")
        assert repr(expr) == "a"
        assert expr.is_variable()

    def test_indexed_variable_expression(self):
        expr = VariableExpression("a", [
            Expression(terms=[VariableExpression("i"),
                              ConstantExpression(1),
                              ],
                       operators=["+"]),
            Expression(terms=[VariableExpression("j"),
                              ConstantExpression(1),
                              ],
                       operators=["-"]),
        ])
        assert str(expr) == "a[i + 1][j - 1]"

    def test_grouping_expression(self):
        expr = GroupingExpression(terms=[
            VariableExpression("a"),
            VariableExpression("b"),
            VariableExpression("c"),
        ], operators=["+", "-"])
        assert str(expr) == "(a + b - c)"

    def test_function_expression(self):
        expr = FunctionExpression("max", (
            VariableExpression("i"),
            VariableExpression("j")
        ))
        assert expr.name == "max"
        assert str(expr) == "max(i, j)"
    
    def test_unary_expression(self):
        expr = UnaryExpression(VariableExpression("n"), "-")
        assert str(expr) == "-n"

    def test_empty_expression(self):
        with pytest.raises(ValueError):
            _ = Expression([], [])

    def test_wrong_operator_number(self):
        with pytest.raises(ValueError):
            _ = Expression([
                ConstantExpression(1),
                ConstantExpression(2),
            ], [])

    def test_is_sigle_expression(self):
        expr = ConstantExpression(1)
        assert expr.is_single()

    def test_is_not_sigle_expression(self):
        expr = Expression([
            VariableExpression("i"),
            VariableExpression("j"),
        ], ["+"])
        assert not expr.is_single()

    def test_is_simple_variable_expression(self):
        expr = VariableExpression("a")
        assert expr.is_simple()

    def test_is_not_simple_variable_expression(self):
        expr = VariableExpression("a", [VariableExpression("i")])
        assert not expr.is_simple()

    def test_simple_extract_variable_expressions(self):
        expr = GroupingExpression(terms=[
            VariableExpression("a"),
            VariableExpression("b"),
            VariableExpression("c"),
            ConstantExpression(1)
        ], operators=["+", "-", "+"])
        assert list(str(e) for e in extract_variable_expressions(expr)) == [
            "a", "b", "c"
        ]

    def test_extract_variable_expression_with_function(self):
        expr = FunctionExpression("max", (
            VariableExpression("i"),
            VariableExpression("j")
        ))
        assert list(str(e) for e in extract_variable_expressions(expr)) == [
            "i", "j"
        ]

    def test_complex_extract_variable_expressions(self):
        expr = Expression(
            terms=[
                GroupingExpression(terms=[
                    VariableExpression("a"),
                    VariableExpression("b"),
                    VariableExpression("c")
                ], operators=["+", "-"]),
                ConstantExpression(3)
            ],
            operators=["/"]
        )
        assert list(str(e) for e in extract_variable_expressions(expr)) == [
            "a", "b", "c"
        ]
    
    def test_divide_variable_expressions(self):
        expressions = [
            VariableExpression("a"),
            VariableExpression("b"),
            VariableExpression("a", [
                VariableExpression("i"),
                VariableExpression("j")
            ])
        ]
        expr_dict = divide_variable_expressions_by_name(expressions)
        assert list([str(e) for e in expr_dict["a"]]) == ["a", "a[i][j]"]
        assert list([str(e) for e in expr_dict["b"]]) == ["b"]
        assert "c" not in expr_dict
